# -*- coding: utf-8 -*-
"""The mesh module of dolfin"""
from dolfin.mesh import meshes
from dolfin.mesh import ale
from dolfin.mesh import refinement

from .meshes import *
from .ale import *
from .refinement import *

# NOTE: The automatic documentation system in DOLFIN requires to _not_
# NOTE: define classes or functions within this file. Use separate
# NOTE: modules for that purpose.

__all__ = meshes.__all__ + ale.__all__ + refinement.__all__
